/* N-body calculation using Greenspan 1st-order discrete mechanics.    */
/* Incorporates xyz-parallelism, uses domain pointers, combines        */
/* particle and pair domains, diagonal processors store particles      */
/* Spreading and accumulation coded in PARIS, more globals,            */
/* Booleans stored in char fields.                                     */

#include <stdio.hs>
#include <cm/cm.hs>
#include <math.hs>
#include "cmtime.hs"

/* Max number of bodies in the calculation.  N = 64 fits in 16k procs
   N must be a power of 2, 4N^2 <= no. of v. procs.  LLN = log2(N)    */

#define N  64
#define LN  6

int NB;      /* Number of bodies in system, <= N */
int LNB;     /* Largest power of 2 < NB */
int LLNB;    /* log2(LNB) */
float G;     /* Gravitational constant */

/* Particles, and pairs of particles.  Diagonal entries have particle	*/
/* states.  Some variables could be overloaded. 			*/
/* act  - processor is active pair					*/
/* sm   - processor participates in acceleration sum			*/
/* here - processor represents particle					*/
/* sp   - processor participates in spreading                           */
/* smk  - control for spreading, accumulation, when procs change        */
/* self - self-address for manipulation                                 */

domain pair { int i, j, smk;  char act, sm, here, sp, smm;
	      CM_cubeaddr_t self;
	      float x0, x1, v0, v1, m, d0, r0, ac, w;
	      domain pair *r; };

domain pair P[3][N][N];

/* Cm addresses for PARIS code */

CM_memaddr_t smk_addr, sp_addr, smm_addr, self_addr, ac_addr, w_addr;


/* Pairs will be represented symmetrically.  The function setup
   (below) initializes these so that pi, pj point to appropriate
   particle in corresponding coordinate.   */

void pair::setup()
 { int offset, c;  mono int k;
   for ( LNB = 1, k = 0;  LNB < NB;  LNB <<= 1, k++ ) ;
   sp = (i<LNB) && (j < LNB);  LNB >>= 1;  LLNB = k-1;
   self = (CM_cubeaddr_t)this;
   smk_addr = &smk;  smm_addr = &smm;  sp_addr = &sp;
   self_addr = &self;  ac_addr = &ac;  w_addr = &w;
   offset = (this - &P[0][0][0]);
   c = offset/(N*N);  i = (offset/N)%N;  j = offset%N;
   sm = (i < NB) && (j < NB);
   act = sm && (i != j);  here = sm && (i == j);
   r = (c<2) ? this+N*N : this-2*N*N;
   smm = 1;  smk = i^j;  /* xor */
   for ( k = 1; k < NB; k <<= 1 )
    if (smm && (k & smk)) { smk = k;  smm = 0; }
   if ( !sp ) smk = 0;
 }


void prsmk()
 { int u, v;
   [domain pair].
   for ( u = 0; u < 2*LNB; u++)
    { for ( v = 0; v < 2*LNB; v++) printf("%2d ", P[0][u][v].smk);
      printf("\n"); }
 }

void pair::prpflt(x)
 float x;
 { mono int u, v;
   for ( u = 0; u < 2*LNB; u++)
    { for ( v = 0; v < 2*LNB; v++) printf("%9.2e ", P[0][u][v].x);
      printf("\n"); }
 }

void pair::prpint(x)
 int x;
 { mono int u, v;
   for ( u = 0; u < 2*LNB; u++)
    { for ( v = 0; v < 2*LNB; v++) printf("%8x ", P[0][u][v].x);
      printf("\n"); }
 }

/* Load inital positions, velocities, also masses, gravitational const. */

void init()
 { [domain pair].
    if (here) 
     { x0 = 17.5;  x1 = -23.37;  v0 = 5.14;  m = 25.6;  d0 = 14.0;
       r0 = 3.6;  ac = 4.76; }
 }

void pair::check_ac1(r1)
 float r1;
 { printf("For P[0,5]:\nr1 = %9.5f r0 = %9.5f\n",
                         P[0][0][5].r1, P[0][0][5].r0);
   printf("acx = %9.5f  acy = %9.5f  acz = %9.5f\n",
	  P[0][0][5].ac, P[1][0][5].ac, P[2][0][5].ac);
 }

void pair::check_ac0()
 { printf("For P[0,5]:\nm = %9.3e r0 = %9.5f\n", P[0][0][5].m, P[0][0][5].r0);
   printf("acx = %9.5f  acy = %9.5f  acz = %9.5f\n",
	  P[0][0][5].ac, P[1][0][5].ac, P[2][0][5].ac);
 }

/* smk[i,j] is a power of 2 corresponding to the stage at which proc[i,j]
   sends it's acceleration to partner and drops out of accumulation. It
   also corresponds to the stage at which proc[i,j] begins to send values
   during spreading operations.                                          */

/* Spread xv from diagonal along i-axis.  Leaves result in w. */
/* May spread to power of 2 just larger than NB.              */

void pair::spreadi(xv)
  float xv;
  { mono int k, a, b;  mono CM_memaddr_t xv_addr;
    xv_addr = &xv;  smm = here;
    /* for ( k = 2*LNB, b = LNB*N;  k > 1;  k >>= 1, b >>= 1 ) */
    /*  if ( smm || ( k == smk) )                              */
    /*   { smm = 1;  ((domain pair *)(th^b))->xv = xv; }       */
    a = (int)smk_addr + LLNB + 1;  b = (int)self_addr + LN + LLNB;
    printf("self_addr = %11o  smk_addr = %11o\n",
              (int)self_addr, (int)smk_addr);
    printf("        b = %11o         a = %11o\n", b, a);
    for ( k = LLNB+1;  k > 0;  k--, a--, b-- )
    { CM_move_always(CM_context_flag, smm_addr, 1);
      CM_logior_always(CM_context_flag, (CM_memaddr_t)a, 1);
      CM_move_constant(smm_addr, 1, 1);
      CM_lognot((CM_memaddr_t)b, (CM_memaddr_t)b, 1);
      CM_move(w_addr, xv_addr, 32);
      CM_send(xv_addr, self_addr, w_addr, 32, CM_NOLIMIT);
      CM_lognot((CM_memaddr_t)b, (CM_memaddr_t)b, 1);
      CM_move_always(CM_context_flag, sp_addr, 1);
     }
    w = xv;
  }

/* Spread xv from diagonal along j-axis.  Leaves result in w.  */
/* May spread to power of 2 just larger than NB.               */

void pair::spreadj(xv)
  float xv;
  { mono int k, a, b;  mono CM_memaddr_t xv_addr;
    xv_addr = &xv;  smm = here;
    /* for ( k = 2*LNB; k > 1; )                                   */
    /*   if ( smm || (k == smk) )                                  */
    /*   { smm = 1;  k >>= 1;  ((domain pair *)(th^k))->xv = xv; } */
    a = (int)smk_addr + LLNB + 1;  b = (int)self_addr + LLNB;
    for ( k = LLNB+1;  k > 0;  k--, a--, b-- )
    { CM_move_always(CM_context_flag, smm_addr, 1);
      CM_logior_always(CM_context_flag, (CM_memaddr_t)a, 1);
      CM_move_constant(smm_addr, 1, 1);
      CM_lognot((CM_memaddr_t)b, (CM_memaddr_t)b, 1);
      CM_move(w_addr, xv_addr, 32);
      CM_send(xv_addr, self_addr, w_addr, 32, CM_NOLIMIT);
      CM_lognot((CM_memaddr_t)b, (CM_memaddr_t)b, 1);
      CM_move_always(CM_context_flag, sp_addr, 1);
     }
    w = xv;
  }

/* Sum accelerations.  Called in context sp */

void pair:: sumac()
 { mono int k, a, b;
   smm = 1;  w = 0.0;
   a = (int)self_addr;  b = (int)smk_addr;
    printf("self_addr = %11o  smk_addr = %11o\n",
              (int)self_addr, (int)smk_addr);
    printf("        b = %11o         b = %11o\n", b, a);
   /* Initial context is sp */
   for ( k = 0;  k < 1 + LLNB;  k++, a++, b++ )
    { CM_logand(CM_context_flag, (CM_memaddr_t)b, 1);  /* Select senders */
      CM_move_constant(smm_addr, 0, 1);                /* Delete senders */
      CM_lognot((CM_memaddr_t)a, (CM_memaddr_t)a, 1);  /* Target address */
      CM_send(w_addr, self_addr, ac_addr, 32, CM_NOLIMIT);
      CM_lognot((CM_memaddr_t)a, (CM_memaddr_t)a, 1);  /* Restore address */
      CM_move_always(CM_context_flag, smm_addr, 1);    /* Select receivers */
      ac += w;
     }
   CM_move_always(CM_context_flag, sp_addr, 1);
 }

/* Commands are:
   r - run (prompts for parameters)
   p - print smk
   q - quit                         */

void main()
 { int ns, np, nl, nv, k, vl;  float dt, eps;
   char c[2];
   while ( printf("nbcmtim> "), scanf("%1s", c), c[0] != 'q' )
    if ( c[0] == 'p' ) prsmk(); else
    {
     /* Performs np integrations, each ns steps of dt */
     printf(
         "Enter NB, no. of evals per step, no. of steps, no. of cycles.\n> ");
     scanf("%d%d%d%d", &NB, &nv, &ns, &np);
     [domain pair].{ setup(); };
     init();
     [domain pair]. if (sp)
       { printf("self\n");  prpint((int)self);  prsmk();
         printf("x0 - i\n");  spreadi(x0);  prpflt(w);
         printf("x1 - j\n");  spreadj(x1);  prpflt(w);
         printf("ac\n");  ac = 0;  if (act) ac = 1.0;
         sumac();  prpflt(ac);  printf("w\n");  prpflt(w);
       }
    }
 }

